# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['entity_gym',
 'entity_gym.env',
 'entity_gym.examples',
 'entity_gym.serialization',
 'entity_gym.tests']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'cloudpickle>=2.0.0,<3.0.0',
 'msgpack-numpy>=0.4.7,<0.5.0',
 'msgpack>=1.0.3,<2.0.0',
 'ragged-buffer>=0.3.9,<0.4.0',
 'tqdm>=4.63.1,<5.0.0']

setup_kwargs = {
    'name': 'entity-gym',
    'version': '0.1.3',
    'description': 'Entity Gym',
    'long_description': '# Entity Gym\n\n[![Actions Status](https://github.com/entity-neural-network/entity-gym/workflows/Checks/badge.svg)](https://github.com/entity-neural-network/entity-gym/actions)\n[![PyPI](https://img.shields.io/pypi/v/entity-gym.svg?style=flat-square)](https://pypi.org/project/entity-gym/)\n[![Documentation Status](https://readthedocs.org/projects/entity-gym/badge/?version=latest&style=flat-square)](https://entity-gym.readthedocs.io/en/latest/?badge=latest)\n[![Discord](https://img.shields.io/discord/913497968701747270?style=flat-square)](https://discord.gg/SjVqhSW4Qf)\n\n\nEntity Gym is an open source Python library that defines an _entity based_ API for reinforcement learning environments.\nEntity Gym extends the standard paradigm of fixed-size observation spaces by allowing observations to contain dynamically-sized lists of entities.\nThis enables a seamless and highly efficient interface with simulators, games, and other complex environments whose state can be naturally expressed as a collection of entities.\n\nThe [ENN-PPO library](https://github.com/entity-neural-network/incubator/tree/main/enn_ppo) can be used to train agents for Entity Gym environments.\n\n## Installation\n\n```\npip install entity-gym\n```\n\n## Usage\n\nYou can find tutorials, guides, and an API reference on the [Entity Gym documentation website](https://entity-gym.readthedocs.io/en/latest/index.html).\n\n## Examples\n\nA number of simple example environments can be found in [entity_gym/examples](https://github.com/entity-neural-network/entity-gym/tree/main/entity_gym/examples). The [ENN-Zoo](https://github.com/entity-neural-network/incubator/tree/main/enn_zoo/enn_zoo) project implements Entity Gym bindings for [Procgen](https://github.com/openai/procgen), [Griddly](https://github.com/Bam4d/Griddly), [MicroRTS](https://github.com/santiontanon/microrts), [VizDoom](https://github.com/mwydmuch/ViZDoom), and [CodeCraft](https://github.com/cswinter/DeepCodeCraft).\n',
    'author': 'Clemens Winter',
    'author_email': 'clemenswinter1@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



import random
from copy import deepcopy
from dataclasses import dataclass
from typing import Dict, List, Mapping, Tuple

import numpy as np

from entity_gym.env import (
    Action,
    ActionSpace,
    CategoricalAction,
    CategoricalActionMask,
    CategoricalActionSpace,
    Entity,
    Environment,
    Observation,
    ObsSpace,
)


@dataclass
class Snake:
    color: int
    segments: List[Tuple[int, int]]


@dataclass
class Food:
    color: int
    position: Tuple[int, int]


class MultiSnake(Environment):
    """
    Turn-based version of Snake with multiple snakes.
    Each snake has a different color.
    For each snake, Food of that color is placed randomly on the board.
    Snakes can only eat Food of their color.
    When a snake eats Food of the same color, it grows by one unit.
    When a snake grows and it's length was less than 11, the player receives a reward of 0.1 / num_snakes.
    The game ends when a snake collides with another snake, runs into a wall, eats Food of another color, or all snakes reach a length of 11.
    """

    def __init__(
        self,
        board_size: int = 10,
        num_snakes: int = 2,
        num_players: int = 1,
        max_snake_length: int = 11,
        max_steps: int = 180,
    ):
        """
        :param num_players: number of players
        :param board_size: size of the board
        :param num_snakes: number of snakes per player
        """
        assert num_snakes < 10, f"num_snakes must be less than 10, got {num_snakes}"
        self.board_size = board_size
        self.num_snakes = num_snakes
        self.num_players = num_players
        self.max_snake_length = max_snake_length
        self.snakes: List[Snake] = []
        self.food: List[Food] = []
        self.game_over = False
        self.last_scores = [0] * self.num_players
        self.scores = [0] * self.num_players
        self.step = 0
        self.max_steps = max_steps

    def obs_space(cls) -> ObsSpace:
        return ObsSpace(
            global_features=["step"],
            entities={
                "SnakeHead": Entity(["x", "y", "color"]),
                "SnakeBody": Entity(["x", "y", "color"]),
                "Food": Entity(["x", "y", "color"]),
            },
        )

    def action_space(cls) -> Dict[str, ActionSpace]:
        return {
            "move": CategoricalActionSpace(["up", "down", "left", "right"]),
        }

    def _spawn_snake(self, color: int) -> None:
        while True:
            x = random.randint(0, self.board_size - 1)
            y = random.randint(0, self.board_size - 1)
            if any(
                (x, y) == (sx, sy) for snake in self.snakes for sx, sy in snake.segments
            ):
                continue
            self.snakes.append(Snake(color, [(x, y)]))
            break

    def _spawn_food(self, color: int) -> None:
        while True:
            x = random.randint(0, self.board_size - 1)
            y = random.randint(0, self.board_size - 1)
            if any((x, y) == (f.position[0], f.position[1]) for f in self.food) or any(
                (x, y) == (sx, sy) for snake in self.snakes for sx, sy in snake.segments
            ):
                continue
            self.food.append(Food(color, (x, y)))
            break

    def reset(self) -> Observation:
        self.snakes = []
        self.food = []
        self.game_over = False
        self.last_scores = [0] * self.num_players
        self.scores = [0] * self.num_players
        self.step = 0
        for i in range(self.num_snakes):
            self._spawn_snake(i)
        for i in range(self.num_snakes):
            self._spawn_food(i)
        return self._observe()

    def act(self, actions: Mapping[str, Action]) -> Observation:
        game_over = False
        self.step += 1
        move_action = actions["move"]
        self.last_scores = deepcopy(self.scores)
        food_to_spawn = []
        assert isinstance(move_action, CategoricalAction)
        for id, move in zip(move_action.actors, move_action.indices):
            snake = self.snakes[id]
            x, y = snake.segments[-1]
            if move == 0:
                y += 1
            elif move == 1:
                y -= 1
            elif move == 2:
                x -= 1
            elif move == 3:
                x += 1
            if x < 0 or x >= self.board_size or y < 0 or y >= self.board_size:
                game_over = True
            if any((x, y) == (sx, sy) for s in self.snakes for sx, sy in s.segments):
                game_over = True
            ate_food = False
            snake.segments.append((x, y))
            for i in range(len(self.food)):
                if self.food[i].position == (x, y):
                    if self.food[i].color != snake.color:
                        game_over = True
                    elif len(snake.segments) <= self.max_snake_length:
                        ate_food = True
                        self.scores[id // self.num_snakes] += (
                            1.0 / (self.max_snake_length - 1) / self.num_snakes
                        )
                    self.food.pop(i)
                    # Don't spawn food immediately since it might spawn in front of another snake that hasn't moved yet
                    food_to_spawn.append(snake.color)
                    break
            if not ate_food:
                snake.segments = snake.segments[1:]
        for player in range(self.num_players):
            snakes_per_player = self.num_snakes // self.num_players
            if all(
                len(s.segments) >= self.max_snake_length
                for s in self.snakes[
                    player * snakes_per_player : (player + 1) * snakes_per_player
                ]
            ):
                game_over = True
        for color in food_to_spawn:
            self._spawn_food(color)
        if self.step >= self.max_steps:
            game_over = True
        return self._observe(done=game_over)

    def _observe(self, done: bool = False, player: int = 0) -> Observation:
        color_offset = player * (self.num_snakes // self.num_players)

        def cycle_color(color: int) -> int:
            return (color - color_offset) % self.num_snakes

        return Observation(
            global_features=[self.step],
            features={
                "SnakeHead": np.array(
                    [
                        [
                            s.segments[-1][0],
                            s.segments[-1][1],
                            cycle_color(s.color),
                        ]
                        for s in self.snakes
                    ],
                    dtype=np.float32,
                ),
                "SnakeBody": np.array(
                    [
                        [sx, sy, cycle_color(snake.color)]
                        for snake in self.snakes
                        for sx, sy in snake.segments[:-1]
                    ],
                    dtype=np.float32,
                ).reshape(-1, 3),
                "Food": np.array(
                    [
                        [
                            f.position[0],
                            f.position[1],
                            cycle_color(f.color),
                        ]
                        for f in self.food
                    ],
                    dtype=np.float32,
                ),
            },
            ids={
                "SnakeHead": list(range(self.num_snakes)),
            },
            actions={
                "move": CategoricalActionMask(
                    actor_types=["SnakeHead"],
                ),
            },
            reward=self.scores[player] - self.last_scores[player],
            done=done,
        )



from dataclasses import dataclass, field
from typing import Dict, List, Mapping

import numpy as np

from entity_gym.env import (
    Action,
    ActionSpace,
    Entity,
    EntityID,
    Environment,
    Observation,
    ObsSpace,
    SelectEntityAction,
    SelectEntityActionMask,
    SelectEntityActionSpace,
)
from entity_gym.env.environment import EntityName


@dataclass
class CherryPick(Environment):
    """
    The CherryPick environment is initialized with a list of 32 cherries of random quality.
    On each timestep, the player can pick up one of the cherries.
    The player receives a reward of the quality of the cherry picked.
    The environment ends after 16 steps.
    The quality of the top 16 cherries is normalized so that the maximum total achievable reward is 1.0.
    """

    num_cherries: int = 32
    cherries: List[float] = field(default_factory=list)
    last_reward: float = 0.0
    step: int = 0

    def obs_space(self) -> ObsSpace:
        return ObsSpace(
            entities={
                "Cherry": Entity(["quality"]),
                "Player": Entity([]),
            }
        )

    def action_space(self) -> Dict[str, ActionSpace]:
        return {"Pick Cherry": SelectEntityActionSpace()}

    def reset(self) -> Observation:
        cherries = [np.random.normal() for _ in range(self.num_cherries)]
        # Normalize so that the sum of the top half is 1.0
        top_half = sorted(cherries, reverse=True)[: self.num_cherries // 2]
        sum_top_half = sum(top_half)
        add = 2.0 * (1.0 - sum_top_half) / self.num_cherries
        self.cherries = [c + add for c in cherries]
        self.last_reward = 0.0
        self.step = 0
        self.total_reward = 0.0
        return self.observe()

    def observe(self) -> Observation:
        done = self.step == self.num_cherries // 2
        ids: Dict[EntityName, List[EntityID]] = {
            "Cherry": [("Cherry", a) for a in range(len(self.cherries))],
            "Player": ["Player"],
        }
        return Observation(
            features={
                "Cherry": np.array(self.cherries, dtype=np.float32).reshape(-1, 1),
                "Player": np.zeros([1, 0], dtype=np.float32),
            },
            ids=ids,
            actions={
                "Pick Cherry": SelectEntityActionMask(
                    actor_ids=["Player"],
                    actee_ids=ids["Cherry"],
                ),
            },
            reward=self.last_reward,
            done=done,
        )

    def act(self, actions: Mapping[str, Action]) -> Observation:
        assert len(actions) == 1, actions
        a = actions["Pick Cherry"]
        assert isinstance(a, SelectEntityAction)
        _, chosen_cherry_idx = a.actees[0]
        self.last_reward = self.cherries.pop(chosen_cherry_idx)
        self.total_reward += self.last_reward
        self.step += 1
        return self.observe()
